/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

#include "general.h"

/**************************************************************************
 * The Heap Sort Class Functions
 */

/*************************
 * Indexed "K-best-sort"
 * Managed by a heap.
 * By R.Collo.
 */

void r_handleIndexedKBestSort(int the_index, int *index, real *data, int K)
{
	real wc, wb1, wb2;
	int i1, i2, j;
  
	wc = data[the_index];
	if(wc >= data[index[0]])
	{
		index[0] = the_index;
		j = 0;
      
		for(;j < K>>1;)
		{
			i1 = (j<<1)+1;
			i2 = (j<<1)+2;
	  
			if(i2 < K)
			{
				wb1 = data[index[i1]];
				wb2 = data[index[i2]];
	      
				if( (wc > wb1) || (wc > wb2) )
				{
					if(wb1 <= wb2)
					{
						index[j] = index[i1];
						index[i1] = the_index;
						j = i1;
					}
					else
					{
						index[j] = index[i2];
						index[i2] = the_index;
						j = i2;
					}
				}
				else
					break;
			}
			else
			{
				if( wc > data[index[i1]] )
				{
					index[j] = index[i1];
					index[i1] = the_index;
				}
				break;
			}
		}
	}
}

/*************************
 * Indexed "K-worst-sort"
 * Managed by a heap.
 * By R.Collo.
 */

void r_handleIndexedKWorstSort(int the_index, int *index, real *data, int K)
{
	real wc, wb1, wb2;
	int i1, i2, j;

	wc = data[the_index];
	if(wc <= data[index[0]])
	{
		index[0] = the_index;
		j = 0;
      
		for(;j < K>>1;)
		{
			i1 = (j<<1)+1;
			i2 = (j<<1)+2;
	  
			if(i2 < K)
			{
				wb1 = data[index[i1]];
				wb2 = data[index[i2]];
	      
				if( (wc < wb1) || (wc < wb2) )
				{
					if(wb1 >= wb2)
					{
						index[j] = index[i1];
						index[i1] = the_index;
						j = i1;
					}
					else
					{
						index[j] = index[i2];
						index[i2] = the_index;
						j = i2;
					}
				}
				else
					break;
			}
			else
			{
				if( wc < data[index[i1]] )
				{
					index[j] = index[i1];
					index[i1] = the_index;
				}
	      
				break;
			}
		}
	}
}


/*******************************
 * Indexed Ascending Heap Sort
 * Adapted from code of
 * "numerical recipies"
 */

void r_indexedAscendingSort(int *index, real *data, int size)
{
	int l = (size >> 1) + 1;
	int ir = size;
	int indxt;
	real z;
	int i, j;
 
	index -= 1;
	for(;;)
	{
		if(l > 1)
			z = data[(indxt=index[--l])];
		else
		{
			z = data[(indxt=index[ir])];
			index[ir] = index[1];
			if(--ir == 1)
			{
				index[1] = indxt;
				break;
			}
		}
      
		i = l;
		j = (l << 1);
		while(j <= ir)
		{
			if( (j < ir) && (data[index[j]] < data[index[j+1]]) )
				j++;
	  
			if(z < data[index[j]])
			{
				index[i] = index[j];
				j += (i=j);
			}
			else
				j = ir + 1;
		}
		index[i] = indxt;
	}
}

/********************************
 * Indexed Descending Heap Sort
 * Adapted from code of
 * "numerical recipies"
 */

void r_indexedDescendingSort(int *index, real *data, int size)
{
	int l = (size >> 1) + 1;
	int ir = size;
	int indxt;
	real z;
	int i, j;
  
	index -= 1;
	for(;;)
	{
		if(l > 1)
			z = data[(indxt=index[--l])];
		else
		{
			z = data[(indxt=index[ir])];
			index[ir] = index[1];
			if(--ir == 1)
			{
				index[1] = indxt;
				break;
			}
		}
      
		i = l;
		j = (l << 1);
		while(j <= ir)
		{
			if( (j < ir) && (data[index[j]] > data[index[j+1]]) )
				j++;
	  
			if(z > data[index[j]])
			{
				index[i] = index[j];
				j += (i=j);
			}
			else
				j = ir + 1;
		}	
		index[i] = indxt;
	}
}



/**************************************************************************
 * La meme chose en << int >> : c'est boeuf
 */

/*************************
 * Indexed "K-best-sort"
 * Managed by a heap.
 * By R.Collo.
 */

void i_handleIndexedKBestSort(int the_index, int *index, int *data, int K)
{
	int wc, wb1, wb2;
	int i1, i2, j;
  
	wc = data[the_index];
	if(wc >= data[index[0]])
	{
		index[0] = the_index;
		j = 0;
      
		for(;j < K>>1;)
		{
			i1 = (j<<1)+1;
			i2 = (j<<1)+2;
	  
			if(i2 < K)
			{
				wb1 = data[index[i1]];
				wb2 = data[index[i2]];
	      
				if( (wc > wb1) || (wc > wb2) )
				{
					if(wb1 <= wb2)
					{
						index[j] = index[i1];
						index[i1] = the_index;
						j = i1;
					}
					else
					{
						index[j] = index[i2];
						index[i2] = the_index;
						j = i2;
					}
				}
				else
					break;
			}
			else
			{
				if( wc > data[index[i1]] )
				{
					index[j] = index[i1];
					index[i1] = the_index;
				}
				break;
			}
		}
	}
}

/****************************
 * Indexed "K-worst-sort"
 * Managed by a heap.
 * By R.Collo.
 */

void i_handleIndexedKWorstSort(int the_index, int *index, int *data, int K)
{
	int wc, wb1, wb2;
	int i1, i2, j;

	wc = data[the_index];
	if(wc <= data[index[0]])
	{
		index[0] = the_index;
		j = 0;
      
		for(;j < K>>1;)
		{
			i1 = (j<<1)+1;
			i2 = (j<<1)+2;
	  
			if(i2 < K)
			{
				wb1 = data[index[i1]];
				wb2 = data[index[i2]];
	      
				if( (wc < wb1) || (wc < wb2) )
				{
					if(wb1 >= wb2)
					{
						index[j] = index[i1];
						index[i1] = the_index;
						j = i1;
					}
					else
					{
						index[j] = index[i2];
						index[i2] = the_index;
						j = i2;
					}
				}
				else
					break;
			}
			else
			{
				if( wc < data[index[i1]] )
				{
					index[j] = index[i1];
					index[i1] = the_index;
				}
	      
				break;
			}
		}
	}
}


/*******************************
 * Indexed Ascending Heap Sort
 * Adapted from code of
 * "numerical recipies"
 */

void i_indexedAscendingSort(int *index, int *data, int size)
{
	int l = (size >> 1) + 1;
	int ir = size;
	int indxt;
	int z;
	int i, j;
 
	index -= 1;
	for(;;)
	{
		if(l > 1)
			z = data[(indxt=index[--l])];
		else
		{
			z = data[(indxt=index[ir])];
			index[ir] = index[1];
			if(--ir == 1)
			{
				index[1] = indxt;
				break;
			}
		}
      
		i = l;
		j = (l << 1);
		while(j <= ir)
		{
			if( (j < ir) && (data[index[j]] < data[index[j+1]]) )
				j++;
	  
			if(z < data[index[j]])
			{
				index[i] = index[j];
				j += (i=j);
			}
			else
				j = ir + 1;
		}
		index[i] = indxt;
	}
}

/********************************
 * Indexed Descending Heap Sort
 * Adapted from code of
 * "numerical recipies"
 */

void i_indexedDescendingSort(int *index, int *data, int size)
{
	int l = (size >> 1) + 1;
	int ir = size;
	int indxt;
	int z;
	int i, j;
  
	index -= 1;
	for(;;)
	{
		if(l > 1)
			z = data[(indxt=index[--l])];
		else
		{
			z = data[(indxt=index[ir])];
			index[ir] = index[1];
			if(--ir == 1)
			{
				index[1] = indxt;
				break;
			}
		}
      
		i = l;
		j = (l << 1);
		while(j <= ir)
		{
			if( (j < ir) && (data[index[j]] > data[index[j+1]]) )
				j++;
	  
			if(z > data[index[j]])
			{
				index[i] = index[j];
				j += (i=j);
			}
			else
				j = ir + 1;
		}	
		index[i] = indxt;
	}
}

