/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

/********************************************************************
 * QCchunker (c) R. COLLOBERT 2000
 *---------------------------------------------------------------ZZZZ
 * Categories :
 *
 *   - Chunk select                                 [XXA]
 *   - Cache handling                               [XXB]
 *      - Initialisations diverse                   [XXC]
 *      - Mises a jour                              [XXD]
 *   - Internal loop utils                          [XXE]
 *      - Handle accel and alpha                    [XXF]
 *      - Check Finish                              [XXG]
 *   - Sub Problem Handling                         [XXH]
 *      - Prepare to launch SP                      [XXI]
 *      - Launch SP                                 [XXJ]
 *   - Shrink the problem                           [XXK]
 *   - Initializations & Boxon                      [XXL]
 *   - Internal loop                                [XXX]
 *---------------------------------------------------------------ZZZZ
 ********************************************************************/


#include "chunker.h"

/********************************************************************
 * [XXA] Chunk select.
 *
 * Require:
 *  - r_buffer [size n_active_var]
 *  - active_var
 *  - class_var
 *  - accel
 *  - Y
 *  - n_active_var
 *
 * Set:
 *  -  n_select_var
 *  -  select_var
 ********************************************************************/


void selectChunk(void)
{
	int *index;
	real *tab;
	int i, j, k;
	int melanie;
	int n_cl1, n_cl2;
	int i_1, i_2;
	int q;
	int *indexh, *indexl;
	int *p1, *p2;

	tab = r_buffer;

	/* Build the sort-criterion */
	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		tab[k] = accel[k] - Y[k];
	}

	/*
	 * Check and handle pathologic case
	 *
	 * In the case of X=card{ x / x in CL1 and CL2} < chunk_size,
	 * a variable could appear two times in the next heuristic choice.
	 * [Rq: X = n_CL1 + n_CL2 - n_active_var]
	 *
	 * So, to avoid those things :
	 */

	if( (n_CL1+n_CL2-n_active_var < chunk_size) && (iteration > 1) )
	{
		/*
		 * << A tout probleme il y a une solution boeuf >>
		 *
		 * We sort on the entire data-set. *ugh*.
		 */
		index = i_buffer;
		for(i = 0; i < n_active_var; i++)
			index[i] = active_var[i];
		r_indexedDescendingSort(index, tab, n_active_var);
      
		/* We get the best and the worst */
		i = 0;
		j = n_active_var-1;
		n_select_var = 0;
	
		while( (n_select_var < chunk_size) && (i < j) )
		{
			melanie = 0;
	    
			do
			{
				k = index[i];
				if(class_var[k] & CL1)
				{
					melanie |= 1;
					select_var[n_select_var] = k;
					i++;
					break;
				}
				i++;
			} while(i < j);
			
			do
			{
				k = index[j];
				if(class_var[k] & CL2)
				{
					melanie |= 2;
					select_var[n_select_var+1] = k; /* --> select_var should have the good size <-- */
					j--;
					break;
				}
				j--;
			} while(i < j);
			
			if(melanie == 3)
				n_select_var += 2;
		}
	}

	else
		/*
		 * The normal case.
		 */
	{
		/* size of each sub-ensemble of the chunk */
		n_cl1 = 0;
		n_cl2 = 0;
      
		/* current index for each class */
		i_1 = 0;
		i_2 = 0;
      
		/* we want q examples for each class [!!! q even !!!] */
		q = chunk_size / 2;

		/***************************************/

		/* indexl  : lowest  var 
		 * indexh  : highest var
		 */
		indexh = select_var;
      
		/* Try to get q examples, class 1 */
		while( (n_cl1 < q) && (i_1 < n_active_var) )
		{
			k = active_var[i_1];
			if(class_var[k] & CL1)
				indexh[n_cl1++] = k;
			
			i_1++;
		}
      
		indexl = indexh+n_cl1;
      
		/* Try to get q examples, class 2 */
		while( (n_cl2 < q) && (i_2 < n_active_var) )
		{
			k = active_var[i_2];
			if(class_var[k] & CL2)
				indexl[n_cl2++] = k;
			
			i_2++;
		}
      
		/* Class High : ascendant order */
		if( (n_cl1 == q) && (i_1 < n_active_var) && (q > 1) )
			r_indexedAscendingSort(indexh, tab, q);
      
		/* Class Low : descendent order */
		if( (n_cl2 == q) && (i_2 < n_active_var) && (q > 1) )
			r_indexedDescendingSort(indexl, tab, q);
      
		/* Classe low
		 * Get the q-lowest
		 */
		if( (n_cl2 == q) && (i_2 < n_active_var) )
		{
			for(i = i_2; i < n_active_var; i++)
			{
				k = active_var[i];
				if(class_var[k] & CL2)
					r_handleIndexedKWorstSort(k, indexl, tab, q);
			}
		}
      
		/* Classe high
		 * Get the q-highest
		 */
		if( (n_cl1 == q) && (i_1 < n_active_var) )
		{
			for(i = i_1; i < n_active_var; i++)
			{
				k = active_var[i];
				if(class_var[k] & CL1)
					r_handleIndexedKBestSort(k, indexh, tab, q);
			}
		}
      

		/****************************************
		 *  If there is too few variables in
		 *  a class...
		 *  The case is so scarse that I don't
		 *  know if it *really* works !
		 */
		if(n_cl1 == n_cl2)
			n_select_var = n_cl1<<1;
		else
		{
			if(n_cl1 < n_cl2)
				n_select_var = n_cl1<<1;
			else
			{
				n_select_var = n_cl2<<1;
				p1 = select_var+n_cl2;
				p2 = select_var+n_cl1;
				for(i = 0; i < n_cl2; i++)
					*p1++ = *p2++;
			}
		}
	}
  
	/****************************************
	 *
	 *  This should *NOT* appear.
	 */ 
	if(n_select_var < 2)
	{
		its_the_end = true;
		if(debug)
			printf("Not enough variables.\n");
		return;
	}
}


/********************************************************************
 *  [XXB] Cache handling.
 ********************************************************************/

/**************************************
 *  [XXC] Initialisations diverses
 **************************************/

void initCache(void)
{
	int i;

	/* Init cache memory */
	n_free_cache = (int) ( ((real)cache_size) / ((real)(n*sizeof(real))) );
	mem_cache = (real *)malloc((n_free_cache*n)*sizeof(real));
	if(!mem_cache)
		fatalError("Out of memory");

	/* Init free pointeurs */
	ptr_free_cache = (real **)malloc(n_free_cache*sizeof(real *));
	if(!ptr_free_cache)
		fatalError("Out of memory");
  
	for(i = 0; i < n_free_cache; i++)
		ptr_free_cache[i] = mem_cache+(i*n);

	/* Init columns pointeurs */
	row_ptr_cache = (real **)malloc(n*sizeof(real *));
	if(!row_ptr_cache)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		row_ptr_cache[i] = NULL;

	/* Init number of use of cache [BOEUF] */
	n_use_cache = (int *)malloc(sizeof(int)*n);
	if(!n_use_cache)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		n_use_cache[i] = 0;

	/* For cache request... */
	query_cache = (int *)malloc(sizeof(int)*n);
	if(!query_cache)
		fatalError("Out of memory");
  
	/* Yo man */
	n_total_cache = n_free_cache;

}


/**************************************
 *  [XXD] Mises a jour
 *  [CALL AFTER chunk_select]
 **************************************/

void checkCache(void)
{
	int n_query;
	int i, j, k, q;
	int n_ok;
	int annick;
	real *zarma, *ether, *ptr;

	/* Is there anything to recalculate ? */
	n_query = 0;
	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		i_need_it[k] = true;
		if(!row_ptr_cache[k])
		{
			query_cache[n_query] = k;
			n_query++;
		}
		else
			n_use_cache[k] = iteration;
	}

	/* Clear out the good number.
	 *  "I'm useful" is the criterion to survive.
	 */
	q = n_query - n_free_cache;
	if(q > 0)
	{
		int *index = i_buffer;
		int i_ok = 0;
		
		/* Pecho the first q examples to sort */
		n_ok = 0;
		while( n_ok < q )
		{
			k = active_var[i_ok];
			if( (n_use_cache[k] != 0) && (!i_need_it[k]) )
				index[n_ok++] = k;
			i_ok++;
		}
	

		/* Descending sort. */
		if( (i_ok < n_active_var) && (q > 1) )
			i_indexedDescendingSort(index, n_use_cache, q);
		
		/* Keep the q worst. */
		if( i_ok < n_active_var )
		{
			for(i = i_ok; i < n_active_var; i++)
			{
				k = active_var[i];
				if( (n_use_cache[k] != 0) && (!i_need_it[k]) )
					i_handleIndexedKWorstSort(k, index, n_use_cache, q);
			}
		}
		
		/* Ok. Now we exterminate. */
		for(i = 0; i < q; i++)
		{
			k = index[i];
			ptr_free_cache[n_free_cache] = row_ptr_cache[k];
			row_ptr_cache[k] = NULL;
			n_use_cache[k] = 0;
			n_free_cache++;
		}
		
	}


	/* Now, we fill what we need. */
	for(i = 0; i < n_query; i++)
	{
		k = query_cache[i];

		row_ptr_cache[k] = ptr_free_cache[n_free_cache-1];
		n_use_cache[k] = iteration;
		n_free_cache--;

		/*********------ WARNING : this part is *ugly* -------*/
		ether = points+(k*width);
		ptr = row_ptr_cache[k];
		
		if(we_have_yet_shrink)
		{
			   for(j = 0; j < n_active_var; j++)
			   {
			      annick = active_var[j];
			      ptr[annick] = (*kernel)((points+annick*width), ether);
			   }
		}
		else
		{
			zarma = points;
			for(j = 0; j < n_active_var; j++)
			{
				*ptr++ = (*kernel)(zarma, ether);
				zarma += width;
			}
		}
		/*********--------------------------------------------*/
	}
	

	/* Clear I_need_it' */
	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		i_need_it[k] = false;
	}
}

/********************************************************************
 *  Internal loop utils [XXE]
 ********************************************************************/

/**************************************
 *  [XXF] Handle accel and alpha
 **************************************/
void handleAccelAndAlpha(void)
{
	int i, j;
	real z;
	int k1, k2;
	real *row_ptr;
	real alph, y;

	for(j = 0; j < n_select_var; j++)
	{
		k1 = select_var[j];
		z = alpha_new[j] - alpha[k1];

		/*------------------ Check the class ------*/
		if(class_var[k1] & CL1)
			n_CL1--;
		if(class_var[k1] & CL2)
			n_CL2--;

		alpha[k1] += z;
		alph = alpha[k1];
		y = Y[k1];
      
		if(alph < eps)
		{
			if(y < 0)
			{
				class_var[k1] = CL1;
				n_CL1++;
			}
			else
			{
				class_var[k1] = CL2;
				n_CL2++;
			}
		}
		else
		{
			if(alph > Ceps)
			{
				if(y < 0)
				{
					class_var[k1] = CL2;
					n_CL2++;
				}
				else
				{
					class_var[k1] = CL1;
					n_CL1++;
				}
			}
			else
			{
				class_var[k1] = CL12;
				n_CL1++;
				n_CL2++;
			}
		}
		/*-----------------------------------------*/

		z *= y;
      
		row_ptr = row_ptr_cache[k1];

		for(i = 0; i < n_active_var; i++)
		{
			k2 = active_var[i];
			accel[k2] += z*row_ptr[k2];
		} 
	}
}

/**************************************
 *  [XXG] Check Finish
 **************************************/


bool checkFinish(void)
{
	real erreur, sum, z;
	int i, k, nb;
	erreur = 0;

	/* Estimate "b" */
	nb = 0;
	sum = 0;
	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		if(class_var[k] == CL12)
		{
			sum += Y[k] - accel[k];
			nb++;
		}
	}
	
	if(nb)
		lb = sum / ((real)nb);

	/* Is it finished ? And check shrink. */
	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		z = Y[k]*(accel[k] + lb);
		if( (alpha[k] > eps) && (z > 1) )
		{
			if( (z-1) > erreur )
				erreur = z-1;
		}
		else
		{
			if( (alpha[k] < Ceps) && (z < 1) )
			{
				if( (1-z) > erreur )
					erreur = 1-z;
			}
		}

		
		if(class_var[k] == CL12)
			not_at_bound_at[k]=iteration;
		else
		{
			if( (alpha[k] < eps) && (z < (1+eps)) )
				not_at_bound_at[k]=iteration;	  
			else
			{
				if( (alpha[k] > Ceps) && (z > (1-eps)) )
					not_at_bound_at[k]=iteration;	  
			}
		}
	}
	

	if(debug)
	{
		if(!(iteration % 1000))
			printf("    Current maximal error = %f\n", erreur);
	}

	if(erreur > precision)
		return(false);
	else
		return(true);
}



/********************************************************************
 *  [XXH] Sub Problem Handling
 ********************************************************************/

/**************************************
 *  [XXI] Prepare to launch SP
 **************************************/

void prepareToLaunchSubProblem(void)
{
	real *ptr_K, *ptr_W;
	real *row_ptr;
	int i, j, k;
	real *bp;
	real z;

/*	K_SP = new Mat(n_select_var, n_select_var, K_SP_data); */

	/* Fill the matrix of the sub-problem */
	
	/* Rq: Size of the matrix of the sub-problem : n_select_var*n_select_var
	 *  Fill the lower triangular matrix
	 */
	ptr_K = K_SP;
	for(i = 0; i < n_select_var; i++)
	{
		row_ptr = row_ptr_cache[select_var[i]];
		for(j = 0; j < i; j++)
			*ptr_K++ = row_ptr[select_var[j]];
		
		*ptr_K = row_ptr[select_var[i]];
		ptr_K += (n_select_var-i);
	}
    
	/* Fill the upper triangular matrix */
	ptr_K = K_SP+1;
	for(i = 1; i < n_select_var; i++)
	{
		ptr_W = K_SP+(n_select_var*i)+(i-1);
		
		for(j = i; j < n_select_var; j++)
		{
			*ptr_K++ = *ptr_W;
			ptr_W += n_select_var;
		}
		ptr_K += (i+1);
	}


	/* And now, the input vector of the sub-problem */

	bp = b_SP;
	for(i = 0; i < n_select_var; i++)
		*bp++ = 0;
	
	/* "Samantha" loop */
	for(j = 0; j < n_select_var; j++)
	{
		bp = b_SP;
		k = select_var[j];
		row_ptr = row_ptr_cache[k];
		z = alpha[k] * Y[k];
		
		for(i = 0; i < n_select_var; i++)
			*bp++ -= ( z*row_ptr[select_var[i]] );
	}
	
	bp = b_SP;    
	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		*bp = - (*bp + accel[k])*Y[k]+1 ;
		bp++;
	}
  
	/* Input alpha [alpha_new] and Y [y_SP] of the sub-problem */
	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		alpha_new[i] = alpha[k];
		Y_SP[i] = Y[k];
	}
}


/**************************************
 *  [XXJ] Launch SP
 **************************************/

/* void launchSubProblem(void) */
/* { */
/* 	i_opti->setK(*K_SP); */
/* 	i_opti->setb(b_SP); */
/* 	i_opti->setY(Y_SP); */
/* 	i_opti->optimize(alpha_new); */

/* } */


/*******************************************************************
 *  [XXK] Skrink the problem
 ********************************************************************/

void shrink(void)
{
	int nb = 0;
	int *to_shrink;
	int i, j, k;

	/* What we shrink ? */
	to_shrink = i_buffer;
	for(i = 0; i < n_active_var; i++)
	{
		if(iteration - not_at_bound_at[active_var[i]] > iter_min_to_shrink)
			to_shrink[nb++] = active_var[i];
	}

	/* Shrink if necessary... */
	if(nb > n_min_to_shrink)
	{
		we_have_yet_shrink = true;

		for(i = 0; i < nb; i++)
		{
			k = to_shrink[i];
			/* Desactivate : */
			i_need_it[k] = true;
			/* Check cache : */
			n_use_cache[k] = 0;
			if(row_ptr_cache[k])
			{
				ptr_free_cache[n_free_cache++] = row_ptr_cache[k];
				row_ptr_cache[k] = NULL;
			}
		}

		/* Another loop to update active variables */
		{
			j = 0;
			for(i = 0; i < n_active_var; i++)
			{
				k = active_var[i];
				if(i_need_it[k])
				{
					if(alpha[k] < eps)
						nulls[n_nulls++] = k;
					else
						bsv[n_bsv++] = k;

					/*------------ Class handle -----*/
					if(class_var[k] & CL1)
						n_CL1--;
					if(class_var[k] & CL2)
						n_CL2--;
					/*-------------------------------*/
				}
				else
					active_var[j++] = k;
			}
			n_active_var = j;
			n_min_to_shrink = n_active_var / proportion_to_shrink;
			if(n_min_to_shrink < chunk_size)
				n_min_to_shrink = chunk_size;
		}
	}
}



/*******************************************************************
 *  [XXL] Initialization & Boxon
 ********************************************************************/

/*-----------------------------------------------------------*/

void initAccel(void)
{
	int i;

	accel = (real *)malloc(n*sizeof(real));
	if(!accel)
		fatalError("Out of memory");

	/* Rq: when we start, n_active_var is maximal */
	for(i = 0; i < n_active_var; i++)
		accel[i] = 0;
}

void initAllTheBigBoxon(void)
{
	int i;

	/* Ho ho */
	n_active_var = n;

	/* Check chunk_size even */
	if( (chunk_size % 2) )
		fatalError("The chunk size *MUST* be even");

	/* Check cache size */
	if( (unsigned int)cache_size < (chunk_size*n*sizeof(real)) )
		fatalError("The cache should be able to contain at least <chunk_size> columns.");

	n_min_to_shrink = n / proportion_to_shrink;
	if(n_min_to_shrink < chunk_size)
		n_min_to_shrink = chunk_size;

	/* Euh... */
	if(!i_opti_is_set)
		fatalError("where is the internal optimizer, boy ????");

/*	i_opti->setC(C);
  	i_opti->setKernel(*kernel);
  	i_opti->setPrecision(i_precision);
*/
	/* Mem Alloc */
	active_var      = (int *)malloc(n*sizeof(int));
	class_var       = (int *)malloc(n*sizeof(int));
	select_var      = (int *)malloc((n+1)*sizeof(int));
	i_buffer        = (int *)malloc(n*sizeof(int));
	not_at_bound_at = (int *)malloc(n*sizeof(int));
	if( ! (active_var && class_var && select_var && i_buffer && not_at_bound_at) )
		fatalError("Out of memory");

	r_buffer  = (real *)malloc(n*sizeof(real));
	K_SP = (real *)malloc(chunk_size*chunk_size*sizeof(real));

	if( ! (K_SP && r_buffer) )
		fatalError("Out of memory");
/*	K_SP = NULL; */

	alpha_new = (real *)malloc(chunk_size*sizeof(real));
	b_SP = (real *)malloc(chunk_size*sizeof(real));
	Y_SP = (real *)malloc(chunk_size*sizeof(real));
	if( ! (alpha_new && b_SP && Y_SP) )
		fatalError("Out of memory");

	
	i_need_it = (bool *)malloc(n*sizeof(bool));
	if( !i_need_it )
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		i_need_it[i] = false;

	bsv = (int *)malloc(n*sizeof(int));
	nulls = (int *)malloc(n*sizeof(int));
	if( ! (bsv && nulls) )
		fatalError("Out of memory");

	n_bsv = 0;
	n_nulls = 0;

	initAccel();
	initCache();

	/* OKAY ???? !!!! Start the machine... */
	n_CL1 = 0;
	n_CL2 = 0;
	for(i = 0; i < n; i++)
	{
		alpha[i] = 0;
		active_var[i] = i;
		not_at_bound_at[i] = 0;
      
		if(Y[i] < 0)
		{
			class_var[i] = CL1;
			n_CL1++;
		}
		else
		{
			class_var[i] = CL2;
			n_CL2++;
		}
	}

	lb = 0;
	iteration = 1;
	Ceps = C - eps;
}

void wipeOutTheBigBoxon(void)
{
	free(active_var);
	free(class_var);
	free(accel);
	free(select_var);
	free(r_buffer);
	free(i_buffer);
	free(K_SP);
/*	if(K_SP) */
/*		delete K_SP; */
	free(i_need_it);
	free(not_at_bound_at);

	free(mem_cache);
	free(ptr_free_cache);
	free(row_ptr_cache);
	free(n_use_cache);
	free(query_cache);
	free(bsv);
	free(nulls);
}


/*******************************************************************
 *  [XXX] Internal Loop
 ********************************************************************/

void solve(real *the_alpha, int the_n, int the_width)
{
    
	/******** Time */
#ifdef I_WANT_TIME
	long run_deb, run_end;
	run_deb = getRuntime();
#endif
	/************************/

	its_the_end = false;
	we_have_yet_shrink = false;

	/* Yeah */
	alpha = the_alpha;
/*	n = alpha.length; */
	n = the_n;
	width = the_width;

	/* Init all */
	initAllTheBigBoxon();

	/* Go !
	 */
	do
	{
		if(iteration > iter_min_to_shrink)
		{
			if(!(iteration % 10))
				shrink();
		}

		selectChunk();

	   if(!its_the_end)
		{
			checkCache();
			prepareToLaunchSubProblem();
			(*launchSubProblem)();
			handleAccelAndAlpha();

			iteration++;

			if(debug)
			{
				if(!(iteration % 1000))
					printf("--> Iteration = %d .... Number of actives variables = %d\n", iteration, n_active_var);
			}
	  
			its_the_end |= checkFinish();
		}

	} while(!its_the_end);


	/******** Time */
#ifdef I_WANT_TIME
	if(debug)
	{
		run_end = getRuntime();	
		printf("\n\nTime in CPU-seconds = %f\n\n", (double)(run_end-run_deb)/((double)CLK_TCK));
	}
#endif
	/***************/

	/* Erase all */
	wipeOutTheBigBoxon();

	return;

}
