/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

/********************************************************************
 * QRchunker (c) R. COLLOBERT 2000
 *---------------------------------------------------------------ZZZZ
 * Categories :
 *
 *   - Chunk select                                 [XXA]
 *   - Cache handling                               [XXB]
 *      - Initialisations diverse                   [XXC]
 *      - Mises a jour                              [XXD]
 *   - Internal loop utils                          [XXE]
 *      - Handle accel and alpha                    [XXF]
 *      - Check Finish                              [XXG]
 *   - Sub Problem Handling                         [XXH]
 *      - Prepare to launch SP                      [XXI]
 *      - Launch SP                                 [XXJ]
 *   - Shrink the problem                           [XXK]
 *   - Initializations & Boxon                      [XXL]
 *   - Internal loop                                [XXX]
 *---------------------------------------------------------------ZZZZ
 ********************************************************************/

#include "chunker.h"

#define CL1  1
#define CL2  2
#define CL12 3

/* Chunk Select
 */

void RselectChunk(void)
{
	int *index;
	real *tab;
	int i, j, k;
	int melanie;
	int n_cl1, n_cl2;
	int i_1, i_2;
	int q;
	int *indexh, *indexl;
	int *p1, *p2;

	tab = r_buffer;

	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		if(k < n)
			tab[k] =  accel[k];
		else
			tab[k] = -accel[k];
	}

	if( (n_CL1+n_CL2-n_active_var < chunk_size) && (iteration > 1) )
	{
		index = i_buffer;
		for(i = 0; i < n_active_var; i++)
			index[i] = active_var[i];
		r_indexedDescendingSort(index, tab, n_active_var);
      
		i = 0;
		j = n_active_var-1;
		n_select_var = 0;
		
		while( (n_select_var < chunk_size) && (i < j) )
		{
			melanie = 0;
	    
			do
			{
				k = index[i];
				if(class_var[k] & CL1)
				{
					melanie |= 1;
					select_var[n_select_var] = k;
					i++;
					break;
				}
				i++;
			} while(i < j);
			
			do
			{
				k = index[j];
				if(class_var[k] & CL2)
				{
					melanie |= 2;
					select_var[n_select_var+1] = k;
					j--;
					break;
				}
				j--;
			} while(i < j);
			
			if(melanie == 3)
				n_select_var += 2;
		}
	}
	else
	{
		n_cl1 = 0;
		n_cl2 = 0;
      
		i_1 = 0;
		i_2 = 0;
      
		q = chunk_size / 2;

		indexh = select_var;
      
		while( (n_cl1 < q) && (i_1 < n_active_var) )
		{
			k = active_var[i_1];
			if(class_var[k] & CL1)
				indexh[n_cl1++] = k;
			
			i_1++;
		}
      
		indexl = indexh+n_cl1;
      
		while( (n_cl2 < q) && (i_2 < n_active_var) )
		{
			k = active_var[i_2];
			if(class_var[k] & CL2)
				indexl[n_cl2++] = k;
			
			i_2++;
		}
      
		if( (n_cl1 == q) && (i_1 < n_active_var) && (q > 1) )
			r_indexedAscendingSort(indexh, tab, q);
      
		if( (n_cl2 == q) && (i_2 < n_active_var) && (q > 1) )
			r_indexedDescendingSort(indexl, tab, q);

		if( (n_cl2 == q) && (i_2 < n_active_var) )
		{
			for(i = i_2; i < n_active_var; i++)
			{
				k = active_var[i];
				if(class_var[k] & CL2)
					r_handleIndexedKWorstSort(k, indexl, tab, q);
			}
		}
      
		if( (n_cl1 == q) && (i_1 < n_active_var) )
		{
			for(i = i_1; i < n_active_var; i++)
			{
				k = active_var[i];
				if(class_var[k] & CL1)
					r_handleIndexedKBestSort(k, indexh, tab, q);
			}
		}
      

		if(n_cl1 == n_cl2)
			n_select_var = n_cl1<<1;
		else
		{
			if(n_cl1 < n_cl2)
				n_select_var = n_cl1<<1;
			else
			{
				n_select_var = n_cl2<<1;
				p1 = select_var+n_cl2;
				p2 = select_var+n_cl1;
				for(i = 0; i < n_cl2; i++)
					*p1++ = *p2++;
			}
		}
	}
  
	if(n_select_var < 2)
	{
		its_the_end = true;
		if(debug)
			printf("Not enough variables.\n");
		return;
	}
}


/*
 * Cache handling.
 */

/*
 * Initialisations diverses
 */

void RinitCache(void)
{
	int i;

	n_free_cache = (int) ( ((real)cache_size) / ((real)(n*sizeof(real))) );
	mem_cache = (real *)malloc(sizeof(real)*(n_free_cache*n));
	if(!mem_cache)
		fatalError("Out of memory");

	ptr_free_cache = (real **)malloc(n_free_cache*sizeof(real *));
	if(!ptr_free_cache)
		fatalError("Out of memory");
  
	for(i = 0; i < n_free_cache; i++)
		ptr_free_cache[i] = mem_cache+(i*n);

	row_ptr_cache = (real **)malloc(n*sizeof(real *));
	if(!row_ptr_cache)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		row_ptr_cache[i] = NULL;

	n_use_cache = (int *)malloc(sizeof(int)*n);
	if(!n_use_cache)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		n_use_cache[i] = 0;

	query_cache = (int *)malloc(sizeof(int)*n);;
	if(!query_cache)
		fatalError("Out of memory");
  
	n_total_cache = n_free_cache;
}



void RcheckCache(void)
{
	int n_query;
	int i, j, k, q;
	int n_ok;
	int annick;
	real *zarma, *ether, *ptr;
	int *index;
	int i_ok;

	n_query = 0;
	for(i = 0; i < n_select_var; i++)
	{
		k = (select_var[i]%n);
		if(!i_need_it[k])
		{
			i_need_it[k] = true;
			if(!row_ptr_cache[k])
			{
				query_cache[n_query] = k;
				n_query++;
			}
			else
				n_use_cache[k] = iteration;
		}
	}

	q = n_query - n_free_cache;
	if(q > 0)
	{
		index = i_buffer;
		i_ok = 0;
		
		n_ok = 0;
		while( n_ok < q )
		{
			k = active_mod_var[i_ok];
			if( (n_use_cache[k] != 0) && (!i_need_it[k]) )
				index[n_ok++] = k;
			
			i_ok++;
		}

		if( (i_ok < n_active_mod_var) && (q > 1) )
			i_indexedDescendingSort(index, n_use_cache, q);
		
		if( i_ok < n_active_mod_var )
		{
			for(i = i_ok; i < n_active_mod_var; i++)
			{
				k = active_mod_var[i];
				if( (n_use_cache[k] != 0) && (!i_need_it[k]) )
					i_handleIndexedKWorstSort(k, index, n_use_cache, q);
			}
		}

		for(i = 0; i < q; i++)
		{
			k = index[i];
			ptr_free_cache[n_free_cache] = row_ptr_cache[k];
			row_ptr_cache[k] = NULL;
			n_use_cache[k] = 0;
			n_free_cache++;
		}		
	}

	/* Now, we fill what we need. */
	for(i = 0; i < n_query; i++)
	{
		k = query_cache[i];

		row_ptr_cache[k] = ptr_free_cache[n_free_cache-1];
		n_use_cache[k] = iteration;
		n_free_cache--;

		/*********------ WARNING : this part is *ugly* -------*/
		ether = points+(k*width);
		ptr = row_ptr_cache[k];
		
		if(we_have_yet_shrink)
		{
			   for(j = 0; j < n_active_mod_var; j++)
			   {
			      annick = active_mod_var[j];
			      ptr[annick] = (*kernel)((points+annick*width), ether);
			   }
		}
		else
		{
			zarma = points;
			for(j = 0; j < n_active_mod_var; j++)
			{
				*ptr++ = (*kernel)(zarma, ether);
				zarma += width;
			}
		}
		/*********--------------------------------------------*/
	}
	
	
/* Clear I_need_it' STUPID BOY */
	for(i = 0; i < n_select_var; i++)
		i_need_it[select_var[i]%n] = false;
	
}


void RhandleAccelAndAlpha(void)
{
	real z;
	int k1, k2;
	real *row_ptr;
	real alph;
	int i, j;
	
	for(j = 0; j < n_select_var; j++)
	{
		k1 = select_var[j];
		if(CL_SP[j] > 0)
			z =  alpha_new[j] - alpha[k1];
		else
			z = -alpha_new[j] - alpha[k1];
		
		if(class_var[k1] & CL1)
			n_CL1--;
		if(class_var[k1] & CL2)
			n_CL2--;
		
		alpha[k1] += z;
		alph = alpha[k1];

		if(alph < eps)
		{
			if(k1 < n)
			{
				class_var[k1] = CL2;
				n_CL2++;
			}
			else
			{
				class_var[k1] = CL1;
				n_CL1++;
			}
		}
		else
		{
			if(alph > Ceps)
			{
				if(k1 < n)
				{
					class_var[k1] = CL1;
					n_CL1++;
				}
				else
				{
					class_var[k1] = CL2;
					n_CL2++;
				}
			}
			else
			{
				class_var[k1] = CL12;
				n_CL1++;
				n_CL2++;
			}
		}

		if(k1 < n)
		{
			row_ptr = row_ptr_cache[k1];

			for(i = 0; i < n_active_var; i++)
			{
				k2 = active_var[i];
				if(k2 < n)
					accel[k2] += z*row_ptr[k2];
				else
					accel[k2] -= z*row_ptr[k2-n];
			} 
		}
		else
		{
			row_ptr = row_ptr_cache[k1-n];

			for(i = 0; i < n_active_var; i++)
			{
				k2 = active_var[i];
				if(k2 < n)
					accel[k2] -= z*row_ptr[k2];
				else
					accel[k2] += z*row_ptr[k2-n];
			} 
		}
	}
}

bool RcheckFinish(void)
{
	real erreur = 0;
	int i;
	int k, nb;
	real sum, z;

	nb = 0;
	sum = 0;
	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		if(class_var[k] == CL12)
		{
			if(k < n)
				sum += accel[k];
			else
				sum -= accel[k];
			nb++;
		}
	}

	if(nb)
		lb = sum / ((real)nb);

	for(i = 0; i < n_active_var; i++)
	{
		k = active_var[i];
		if(k < n)
			z = accel[k] - lb;
		else
			z = accel[k] + lb;
		
		if( (alpha[k] > eps) && (z > 0) )
		{
			if( z > erreur )
				erreur = z;
		}
		else
		{
			if( (alpha[k] < Ceps) && (z < 0) )
			{
				if( -z > erreur )
					erreur = -z;
			}
		}

		
		if(class_var[k] == CL12)
			not_at_bound_at[k]=iteration;
		else
		{
			if( (alpha[k] < eps) && (z < eps) )
				not_at_bound_at[k]=iteration;
			else
			{
				if( (alpha[k] > Ceps) && ( -z < eps ) )
					not_at_bound_at[k]=iteration;
			}
		}
	}

	if(debug)
	{
		if(!(iteration % 1000))
			printf("    Current maximal error = %f\n", erreur);
	}

	if(erreur > precision)
		return(false);
	else
		return(true);
}


void RprepareToLaunchSubProblem(void)
{
	int i, j, k;
	real *ptr_K, *ptr_W;
	real *row_ptr;
	real *bp;
	real z;

	ptr_K = K_SP;
	for(i = 0; i < n_select_var; i++)
	{
		row_ptr = row_ptr_cache[select_var[i]%n];
		for(j = 0; j < i; j++)
			*ptr_K++ = row_ptr[select_var[j]%n];
		
		*ptr_K = row_ptr[select_var[i]%n];
		ptr_K += (n_select_var-i);
	}
	
	ptr_K = K_SP+1;
	for(i = 1; i < n_select_var; i++)
	{
		ptr_W = K_SP+(n_select_var*i)+(i-1);
		
		for(j = i; j < n_select_var; j++)
		{
			*ptr_K++ = *ptr_W;
			ptr_W += n_select_var;
		}
		ptr_K += (i+1);
	}

	bp = b_SP;
	for(i = 0; i < n_select_var; i++)
		*bp++ = 0;
	
	for(j = 0; j < n_select_var; j++)
	{
		bp = b_SP;
		k = select_var[j];
		row_ptr = row_ptr_cache[k%n];
		z = alpha[k];
		if(k < n)
			CL_SP[j] = 1;
		else
		{
			CL_SP[j] = -1;
			z = -z;
		}
		
		for(i = 0; i < n_select_var; i++)
			*bp++ += ( z*row_ptr[select_var[i]%n] );
	}
	
	bp = b_SP;
	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		if(k < n)
			*bp++ -= accel[k];
		else
			*bp++ += accel[k];
	}

	for(i = 0; i < n_select_var; i++)
	{
		k = select_var[i];
		alpha_new[i] = alpha[k]*CL_SP[i];
	}
}

void Rshrink(void)
{
	int i, j, k, k1;
	int nb = 0;
	int *to_shrink;

	to_shrink = i_buffer;
	for(i = 0; i < n_active_var; i++)
	{
		if(iteration - not_at_bound_at[active_var[i]] > iter_min_to_shrink)
			to_shrink[nb++] = active_var[i];
	}

	if(nb > n_min_to_shrink)
	{
		we_have_yet_shrink = true;

		for(i = 0; i < nb; i++)
		{	 
			k = to_shrink[i];
			i_need_it[k] = true;

			if(k < n)
				atomise_cache[k]   |= 1;
			else
				atomise_cache[k-n] |= 2;
		}

		j = 0;
		n_active_mod_var = 0;

		for(i = 0; i < n_active_var; i++)
		{
			k = active_var[i];
			if(i_need_it[k])
			{
				if(alpha[k] < eps)
					nulls[n_nulls++] = k;
				else
					bsv[n_bsv++] = k;
				
				if(class_var[k] & CL1)
					n_CL1--;
				if(class_var[k] & CL2)
					n_CL2--;
				
				k1 = k%n;
				if(atomise_cache[k1] == 3)
				{
					n_use_cache[k1] = 0;
					if(row_ptr_cache[k1])
					{
						ptr_free_cache[n_free_cache++] = row_ptr_cache[k1];
						row_ptr_cache[k1] = NULL;
					}
				}
			}
			else
			{		    
				active_var[j++] = k;
				
				k1 = k%n;
				if(!is_registered[k1])
				{
					is_registered[k1] = true;
					active_mod_var[n_active_mod_var++] = k1;
				}
			}
		}
	    
		for(i = 0; i < n_active_mod_var; i++)
			is_registered[active_mod_var[i]] = false;
		
		n_active_var = j;
		n_min_to_shrink = n_active_var / proportion_to_shrink;
		if(n_min_to_shrink < chunk_size)
			n_min_to_shrink = chunk_size;
	}

	for(i = 0; i < nb; i++)
		i_need_it[to_shrink[i]] = false;
	
}

void RinitAccel(void)
{
	int i;

	accel = (real *)malloc(sizeof(real)*2*n);
	if(!accel)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
	{	
		accel[i  ] =  Y[i] + epsilon;
		accel[i+n] = -Y[i] + epsilon;
	}
}

void RinitAllTheBigBoxon(void)
{
	int i;

	n_active_var = 2*n;

	if( (chunk_size % 2) )
		fatalError("The chunk size *MUST* be even");

	if( (unsigned int)cache_size < (chunk_size*n*sizeof(real)) )
		fatalError("The cache should be able to contain at least <chunk_size> columns.");

	n_min_to_shrink = 2*n / proportion_to_shrink;
	if(n_min_to_shrink < chunk_size)
		n_min_to_shrink = chunk_size;

	if(!i_opti_is_set)
		fatalError("where is the internal optimizer, boy ????");

	active_var      = (int *)malloc(sizeof(int)*2*n);
	active_mod_var  = (int *)malloc(sizeof(int)*n);
	class_var       = (int *)malloc(sizeof(int)*2*n);
	select_var      = (int *)malloc(sizeof(int)*(2*n+1));
	i_buffer        = (int *)malloc(sizeof(int)*2*n);
	not_at_bound_at = (int *)malloc(sizeof(int)*2*n);
	if( ! (active_var && class_var && select_var && i_buffer && not_at_bound_at) )
		fatalError("Out of memory");

	if( !active_mod_var )
		fatalError("Out of memory");
	
	r_buffer  = (real *)malloc(sizeof(real)*2*n);
	K_SP = (real *)malloc(sizeof(real)*chunk_size*chunk_size);

	if( ! (K_SP && r_buffer) )
		fatalError("Out of memory");


	alpha_new = (real *)malloc(chunk_size*sizeof(real));
	b_SP = (real *)malloc(chunk_size*sizeof(real));
	if( ! (alpha_new && b_SP) )
		fatalError("Out of memory");
	
	i_need_it = (bool *)malloc(sizeof(bool)*2*n);
	if( !i_need_it )
		fatalError("Out of memory");

	is_registered = (bool *)malloc(sizeof(bool)*n);
	if( !is_registered )
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		is_registered[i] = false;

	for(i = 0; i < 2*n; i++)
		i_need_it[i] = false;

	bsv = (int *)malloc(sizeof(int)*2*n);
	nulls = (int *)malloc(sizeof(int)*2*n);
	if( ! (bsv && nulls) )
		fatalError("Out of memory");

	n_bsv = 0;
	n_nulls = 0;

	RinitAccel();
	RinitCache();

	CL_SP = (real *)malloc(sizeof(real)*chunk_size);
	if(!CL_SP)
		fatalError("Out of memory");

	atomise_cache = (int *)malloc(sizeof(int)*n);
	if(!atomise_cache)
		fatalError("Out of memory");

	for(i = 0; i < n; i++)
		atomise_cache[i] = 0;

	n_CL1 = 0;
	n_CL2 = 0;
	for(i = 0; i < 2*n; i++)
	{
		alpha[i] = 0;
		active_var[i] = i;
		not_at_bound_at[i] = 0;
      
		if(i < n)
		{
			class_var[i] = CL2;
			n_CL2++;
		}
		else
		{
			class_var[i] = CL1;
			n_CL1++;
		}
	}

	for(i = 0; i < n; i++)
		active_mod_var[i] = i;
	n_active_mod_var = n;

	lb = 0;
	iteration = 1;
	Ceps = C - eps;
}

void RwipeOutTheBigBoxon(void)
{
	free(active_var);
	free(class_var);
	free(accel);
	free(select_var);
	free(r_buffer);
	free(i_buffer);
	free(K_SP);
	free(i_need_it);
	free(not_at_bound_at);
	free(CL_SP);

	free(mem_cache);
	free(ptr_free_cache);
	free(row_ptr_cache);
	free(n_use_cache);
	free(query_cache);
	free(bsv);
	free(nulls);
	free(atomise_cache);
	free(is_registered);
}

void Rsolve(real *the_alpha, int the_n, int the_width)
{
#ifdef I_WANT_TIME
	long run_deb, run_end;
	run_deb = getRuntime();
#endif

	its_the_end = false;
	we_have_yet_shrink = false;

	alpha = the_alpha;
	n = the_n;
	width = the_width;

	RinitAllTheBigBoxon();
  
	do
	{
		if(iteration > iter_min_to_shrink)
		{
			if(!(iteration % 10))
				Rshrink();
		}

		RselectChunk();

		if(!its_the_end)
		{
			RcheckCache();
			RprepareToLaunchSubProblem();
			(*launchSubProblem)();
			RhandleAccelAndAlpha();

			iteration++;

			if(debug)
			{
				if(!(iteration % 1000))
					printf("--> Iteration = %d .... Number of actives variables = %d\n", iteration, n_active_var);
			}
	  
			its_the_end |= RcheckFinish();
		}

	} while(!its_the_end);

#ifdef I_WANT_TIME
	if(debug)
	{
		run_end = getRuntime();	
		printf("\n\nTime in CPU-seconds = %f\n\n", (double)(run_end-run_deb)/((double)CLK_TCK));
	}
#endif

	RwipeOutTheBigBoxon();

	return;
}
