/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

#include "svm.h"
#include "kernel.h"

extern char *version;

void error(void)
{
	printf("Error inside. Check your data.\n\n");
	exit(0);
}

void m_error(void)
{
	printf("Memory Error. Buy new RAM or check your data.\n\n");
	exit(0);
}

SVM *initSVM(void)
{
	SVM *svm;

	if( (svm = (SVM *)malloc(sizeof(SVM))) == NULL)
		m_error();

	svm->n_s_v = 0;
	svm->support_vectors = NULL;
	
	svm->points = NULL;
	svm->Y = NULL;

	svm->alpha = NULL;
	svm->n = 0;
	svm->w = 0;

	return(svm);
}

void destroySVM(SVM *svm)
{
	if(svm->support_vectors)
		free(svm->support_vectors);

	if(svm->points)
		free(svm->points);

	if(svm->Y)
		free(svm->Y);

	if(svm->alpha)
		free(svm->alpha);

	free(svm);
}

void createFileAsciiSVM(SVM *svm, char *fc)
{
	FILE *f;
	int n, w;
	int i, j;

	if( (f = fopen(fc, "rt")) == NULL)
		error();

	fscanf(f, "%d", &n);
	fscanf(f, "%d", &w);

	w -= 1;
	svm->n = n;
	svm->w = w;

	if( (svm->points = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (svm->Y = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	if(regression)
	{
		if( (svm->alpha = (real *)malloc(2*n*sizeof(real))) == NULL)
			m_error();
	}
	else
	{
		if( (svm->alpha = (real *)malloc(n*sizeof(real))) == NULL)
			m_error();		
	}
	
	for(i = 0; i < n; i++)
	{
		for(j = 0; j < w; j++)
#ifdef USEDOUBLE
			fscanf(f, "%lf", &svm->points[i*w+j]);
#else
			fscanf(f, "%f", &svm->points[i*w+j]);
#endif

#ifdef USEDOUBLE
		fscanf(f, "%lf", &svm->Y[i]);
#else
		fscanf(f, "%f", &svm->Y[i]);
#endif

		svm->alpha[i] = 0;
		if(regression)
			svm->alpha[i+n] = 0;

	}

	fclose(f);
}


void learnSVM(SVM *svm)
{
	points = svm->points;
	Y = svm->Y;

	/* Yoh Man */
	printf("\nSystem Loaded\n");
	if(regression)
		Rsolve(svm->alpha, svm->n, svm->w);
	else
		solve(svm->alpha, svm->n, svm->w);
		
	svm->b = lb;
	printf("System Atomised\n\n");

}

void createFileBinSVM(SVM *svm, char *fc)
{
	FILE *f;
	int n, w;
	int i;

	if( (f = fopen(fc, "rb")) == NULL)
		error();

	fread(&n, sizeof(int), 1, f);
	fread(&w, sizeof(int), 1, f);

	svm->n = n;
	svm->w = w;

	if( (svm->points = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (svm->Y = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	if(regression)
	{
		if( (svm->alpha = (real *)malloc(2*n*sizeof(real))) == NULL)
			m_error();
	}
	else
	{
		if( (svm->alpha = (real *)malloc(n*sizeof(real))) == NULL)
			m_error();
	}
	
	fread(svm->points, sizeof(real), n*w, f);
	fread(svm->Y, sizeof(real), n, f);

	if(regression)
	{
		for(i = 0; i < 2*n; i++)
			svm->alpha[i] = 0;
	}
	else
	{
		for(i = 0; i < n; i++)
			svm->alpha[i] = 0;
	}

	fclose(f);
}

void checkSVM(SVM *svm)
{
	int i;
	real *ptr;

	printf("\nChecking SVM...\n\n");
	printf("  # examples  : %d\n", svm->n);
	printf("  # dimension : %d\n", svm->w);

	if(!regression)
	{
		ptr = svm->Y;
		for(i = 0; i < svm->n; i++)
		{
			if( !( (*ptr == 1) || (*ptr == -1) ) )
			{
				printf("  # error     : classe must be +1 or -1\n\n");
				exit(0);
			}
			ptr++;
		}
	}

	printf("  # OK\n\n");
	
}

void svSVM(SVM *svm)
{
	int i, k;
	int *ptr;
	real *z;

	if(regression)
	{
		if( (svm->support_vectors = (int *)malloc(sizeof(int)*2*svm->n)) == NULL)
			m_error();
		
		ptr = svm->support_vectors;
		z = svm->alpha;
		
		k = 0;
		for(i = 0; i < 2*svm->n; i++)
		{
			if(*z++ > eps)
				ptr[k++] = i;
		}
	}
	else
	{
		if( (svm->support_vectors = (int *)malloc(sizeof(int)*svm->n)) == NULL)
			m_error();
		
		ptr = svm->support_vectors;
		z = svm->alpha;
		
		k = 0;
		for(i = 0; i < svm->n; i++)
		{
			if(*z++ > eps)
				ptr[k++] = i;
		}
	}

	svm->n_s_v = k;

	printf("\n  # Supports vectors : %d\n\n", svm->n_s_v);
}


void saveFileAsciiSVM(SVM *svm, char *fc)
{
	FILE *f;
	int i, j, k, w;
	int k1;

	if( (f = fopen(fc, "wt")) == NULL)
		error();

	w = svm->w;

	fprintf(f, "%s\n", version);
	fprintf(f, "%d\n", regression);
	fprintf(f, "%d\n", t_kernel);
	fprintf(f, "%d\n", k_poly);
	fprintf(f, "%.8g\n", k_exp);
	fprintf(f, "%.8g\n", k_sigm_s);
	fprintf(f, "%.8g\n", k_sigm_c);
	fprintf(f, "%s\n", k_user);
	fprintf(f, "%d\n", svm->n_s_v);
	fprintf(f, "%d\n", svm->w);
	fprintf(f, "%.8g\n", svm->b);
	
	if(regression)
	{
		for(i = 0; i < svm->n_s_v; i++)
		{
			k = svm->support_vectors[i];
			if(k < n)
				fprintf(f, "%.32g ", -svm->alpha[k]);
			else
				fprintf(f, "%.32g ",  svm->alpha[k]);
			
			k1 = k%n;
			for(j = 0; j < w; j++)
				fprintf(f, "%.8g ", svm->points[k1*w+j]);
			
			fprintf(f, "\n");
		}
	}
	else
	{
		for(i = 0; i < svm->n_s_v; i++)
		{
			k = svm->support_vectors[i];
			fprintf(f, "%.32g ", svm->alpha[k]*svm->Y[k]);
			
			for(j = 0; j < w; j++)
				fprintf(f, "%.8g ", svm->points[k*w+j]);
			
			fprintf(f, "\n");
		}
	}

	fclose(f);
}

void saveFileBinSVM(SVM *svm, char *fc)
{
	FILE *f;
	int i, k, w;
	real z, *ptr;

	if( (f = fopen(fc, "wb")) == NULL)
		error();

	w = svm->w;

	i = strlen(version);
 	fwrite(&i, sizeof(int), 1, f);
	fwrite(version, sizeof(char), i, f);
	fwrite(&regression, sizeof(int), 1, f);
	fwrite(&t_kernel, sizeof(int), 1, f);
	fwrite(&k_poly, sizeof(int), 1, f);
	fwrite(&k_exp, sizeof(real), 1, f);
	fwrite(&k_sigm_s, sizeof(real), 1, f);
	fwrite(&k_sigm_c, sizeof(real), 1, f);
	i = strlen(k_user);
 	fwrite(&i, sizeof(int), 1, f);
 	fwrite(k_user, sizeof(char), i, f);
 	fwrite(&svm->n_s_v, sizeof(int), 1, f);
 	fwrite(&svm->w, sizeof(int), 1, f);
 	fwrite(&svm->b, sizeof(real), 1, f);

	if(regression)
	{
		for(i = 0; i < svm->n_s_v; i++)
		{
			k = (svm->support_vectors[i])%n;
			
			ptr = &svm->points[k*w];
			fwrite(ptr, sizeof(real), w, f);
		}
		
		for(i = 0; i < svm->n_s_v; i++)
		{		
			k = svm->support_vectors[i];
			if(k < n)
				z = -svm->alpha[k];
			else
				z =  svm->alpha[k];

			fwrite(&z, sizeof(real), 1, f);
		}
	}
	else
	{
		for(i = 0; i < svm->n_s_v; i++)
		{
			k = svm->support_vectors[i];
			
			ptr = &svm->points[k*w];
			fwrite(ptr, sizeof(real), w, f);
		}
		
		for(i = 0; i < svm->n_s_v; i++)
		{		
			k = svm->support_vectors[i];
			z = svm->alpha[k]*svm->Y[k];
			fwrite(&z, sizeof(real), 1, f);
		}
	}

	fclose(f);
}
