/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

#include "general.h"
#include "kernel.h"
#include "svm_test_funcs.h"

char *file_model;
char *file_test;
char *file_output;
int mode_model;
int mode_test;
int mode_output;
int W;

void help(void)
{
	printf("***********************************************************************************\n");
	printf("* The Trebolloc Compagnie presents :\n");
	printf("* SVM Test ");
	printf(VERSION);
	printf(" -- A part of SVMTorch\n");
	printf("*\n");
	printf("*\n");
	printf("* Usage : svm_test [options] model_file test_file\n");
	printf("*\n");
	printf("*\n");
	printf("* Arguments :\n");
	printf("*        model_file    -> file with the SVM model\n");
	printf("*        test_file     -> file to with examples to test\n");
	printf("* Input options :\n");
	printf("*        -am           -> ASCII model\n");
	printf("*        -ae           -> ASCII examples\n");
	printf("*        -no           -> No desired-output in test_file\n");
	printf("* Output options :\n");
	printf("*        -oa <file>    -> write in ASCII the SVM output into <file>\n");
	printf("*        -ob <file>    -> write in binary the SVM output into <file>\n");
	printf("*\n");
	printf("*********>> SVMTorch is (c) Ronan Collobert 2000 (IDIAP) <<************************\n\n");

	exit(0);
}

void scan_cmd(int argc, char **argv)
{
	int i = 1;
	int erreur;
	int maxa;

	if(argc < 3)
		help();

	maxa = argc-2;
	while(i < maxa)
	{
		erreur = 1;

		if(!strcmp(argv[i], "-help"))
		{
			help();
			erreur = 0;
		}

		if(!strcmp(argv[i], "-am"))
		{
			mode_model = 1;
			erreur = 0;
		}

		if(!strcmp(argv[i], "-no"))
		{
			W = 1;
			erreur = 0;
		}

		if(!strcmp(argv[i], "-ae"))
		{
			mode_test = 1;
			erreur = 0;
		}

		if(!strcmp(argv[i], "-oa"))
		{
			i++;
			if(i < maxa)
				file_output = argv[i];
			else
				help();

			mode_output = 2;
			erreur = 0;
		}		

		if(!strcmp(argv[i], "-ob"))
		{
			i++;
			if(i < maxa)
				file_output = argv[i];
			else
				help();

			mode_output = 1;
			erreur = 0;
		}		

		if(erreur)
			help();

		i++;

	}

	file_model = argv[argc-2];
	file_test = argv[argc-1];
}

int main(int argc, char **argv)
{
	FILE *ff;
	SVM *svm;
	TEST *test;
	int i, j, nbproch, missclassified, misscl1;
	real z;
	real mae, mse, msse, maxe, stdv;

	/* regression */
	mae = 0;
	mse = 0;
	msse = 0;
	maxe = 0;
	/**************/

	W = 0;
	mode_output = 0;
	file_output = NULL;
   scan_cmd(argc, argv);

	svm  = initSVM();
	test = initTEST();

	printf("\n");
	printf("# Loading SVM...\n");
	if(mode_model)
		createModelFileAsciiSVM(svm, file_model);
	else
		createModelFileBinSVM(svm, file_model);

	if(regression)
		printf("# Regression\n");
	else
		printf("# Classification\n");

	printf("# Loading TEST...\n");

	if(W && (mode_output == 0))
	{
		printf(" ???? What do you want ?? If you want the output of SVM\n");
		printf("      use -oa or -ob\n");
		exit(0);
	}

	if(W)
	{
		if(mode_test)
			loadAsciiWTEST(test, file_test);
		else
			loadBinWTEST(test, file_test);
	}
	else
	{
		if(mode_test)
			loadAsciiTEST(test, file_test);
		else
			loadBinTEST(test, file_test);
	}

	checkModelSVM(svm);
	setKernel();
	if(W)
		checkWTEST(test);
	else
		checkTEST(test);


	if(test->w != svm->w)
	{
		printf("\nDimension must be the same !!!!\n\n");
		exit(0);
	}
	else
		printf("\n# OK\n\n");

	printf("# Testing in Progress...\n\n");
	printf("                    ");
	printf("[________Test________]\n");
	printf("                    [");

	if(regression)
	{
		if(W)
		{
			j = 0;
			nbproch = 0;
			missclassified = 0;
			misscl1 = 0;
			for(i = 0; i < test->n; i++)
			{
				if(i >= nbproch)
				{
					nbproch = ++j * test->n / 20;
					printf("#");
					fflush(stdout);
				}
				
				z = evaluateSVM(svm, &test->ex[test->w*i]);
				test->results[i] = z;
			}
		}
		else
		{
			j = 0;
			nbproch = 0;
			missclassified = 0;
			misscl1 = 0;
			for(i = 0; i < test->n; i++)
			{
				if(i >= nbproch)
				{
					nbproch = ++j * test->n / 20;
					printf("#");
					fflush(stdout);
				}
				
				z = evaluateSVM(svm, &test->ex[test->w*i]);
				test->results[i] = z;
				z -= test->Y[i];

				if(z < 0)
					z = -z;

				if(z > maxe)
					maxe = z;

				mae += z;
				z *= z;
				mse += z;
				z *= z;
				msse += z;
			}
		}
	}
/***************************************************************************
 */

	else
	{
		if(W)
		{
			j = 0;
			nbproch = 0;
			missclassified = 0;
			misscl1 = 0;
			for(i = 0; i < test->n; i++)
			{
				if(i >= nbproch)
				{
					nbproch = ++j * test->n / 20;
					printf("#");
					fflush(stdout);
				}
				
				z = evaluateSVM(svm, &test->ex[test->w*i]);
				test->results[i] = z;
			}
		}
		else
		{
			j = 0;
			nbproch = 0;
			missclassified = 0;
			misscl1 = 0;
			for(i = 0; i < test->n; i++)
			{
				if(i >= nbproch)
				{
					nbproch = ++j * test->n / 20;
					printf("#");
					fflush(stdout);
				}
				
				z = evaluateSVM(svm, &test->ex[test->w*i]);
				test->results[i] = z;
				if(z*test->Y[i] <= 0)
				{
					missclassified++;
					if(test->Y[i] > 0)
						misscl1++;
				}
			}
		}
	}

	printf("]\n");
	fflush(stdout);

	if(!W)
	{
		if(regression)
		{
			z = (real)test->n;
			mae /= z;
			mse /= z;
			msse /= z;
			
			
			printf("\n  # Mean Absolute Error    : %.5f\n", mae);
			stdv = sqrt(mse-mae*mae);
			printf("     -> Standard Deviation : %.5f\n", stdv);
			printf("\n  # Mean Squared Error     : %.5f\n", mse);
			stdv = sqrt(msse-mse*mse);
			printf("     -> Standard Deviation : %.5f\n", stdv);

			printf("\n  # Maximal Error          : %.5f\n", maxe);
			
		}
		else
		{
			z = 100.0 * (real)(missclassified) / ((real)(test->n));
			printf("\n  # Number of misclassified           : %20d [%.2f%%]\n", missclassified, z);
			printf("       -> False positives             : %20d\n", misscl1);
			printf("       -> False negatives             : %20d\n", missclassified-misscl1);
			z = 100.0 * (real)(test->n-missclassified) / ((real)(test->n));
			printf("\n  # Number of correct classifications : %20d [%.2f%%]\n", test->n-missclassified, z);
		}
	}

	printf("\n");

	if(mode_output)
	{
		printf("# Writing results...\n");
		if(mode_output == 1)
		{
			if( (ff = fopen(file_output, "wb")) == NULL)
			{
				printf("\nError in output file.\n");
				exit(0);
			}

			fwrite(test->results, sizeof(real), test->n, ff);
			fclose(ff);
		}
		else
		{
			if( (ff = fopen(file_output, "wt")) == NULL)
			{
				printf("\nError in output file.\n");
				exit(0);
			}

			for(i = 0; i < test->n; i++)
				fprintf(ff, "%.5f\n", test->results[i]);
			
			fclose(ff);
		}
	}

	printf("\n# End.\n\n");

	return(0);
}
