/***********************************************************************/
/*                                                                     */
/*   SVMTorch                                                          */
/*                                                                     */
/*   Author: Ronan Collobert                                           */
/*   Date: 21.07.2000                                                  */
/*                                                                     */
/*   Copyright (c) 2000  IDIAP - All rights reserved                   */
/*                                                                     */
/*   This software is available for non-commercial use only. It must   */
/*   not be modified and distributed without prior permission of the   */
/*   author. The author is not responsible for implications from the   */
/*   use of this software.                                             */
/*                                                                     */
/***********************************************************************/
  

#include "svm_test_funcs.h"
#include "kernel.h"

void error(void)
{
	printf("Error inside. Check your data.\n\n");
	exit(0);
}

void m_error(void)
{
	printf("Memory Error. Buy new RAM or check your data.\n\n");
	exit(0);
}

SVM *initSVM(void)
{
	SVM *svm;

	if( (svm = (SVM *)malloc(sizeof(SVM))) == NULL)
		m_error();

	svm->points = NULL;
	svm->alpha = NULL;
	svm->n = 0;
	svm->w = 0;

	return(svm);
}

TEST *initTEST(void)
{
	TEST *test;

	if( (test = (TEST *)malloc(sizeof(TEST))) == NULL)
		m_error();

	test->ex = NULL;
	test->Y = NULL;
	test->results = NULL;
	test->n = 0;
	test->w = 0;
	
	return(test);
}

void destroySVM(SVM *svm)
{
	if(svm->points)
		free(svm->points);

	if(svm->alpha)
		free(svm->alpha);

	free(svm);
}

void destroyTEST(TEST *test)
{
	if(test->ex)
		free(test->ex);

	if(test->Y)
		free(test->Y);

	if(test->results)
		free(test->results);

	free(test);
}

void checkTEST(TEST *test)
{
	int i;
	real *ptr;

	printf("\n# Checking TEST...\n\n");
	printf("  # examples  : %d\n", test->n);
	printf("  # dimension : %d\n", test->w);

	if(!regression)
	{
		ptr = test->Y;
		for(i = 0; i < test->n; i++)
		{
			if( !( (*ptr == 1) || (*ptr == -1) ) )
				printf("  # warning   : classe must be +1 or -1\n");
			ptr++;
		}
	}

	printf("  # OK\n\n");
	
}

void checkWTEST(TEST *test)
{
	printf("\n# Checking TEST...\n\n");
	printf("  # examples  : %d\n", test->n);
	printf("  # dimension : %d\n", test->w);
	printf("  # OK\n\n");
}

void loadBinTEST(TEST *test, char *fc)
{
	FILE *f;
	int n, w;

	if( (f = fopen(fc, "rb")) == NULL)
		error();

	fread(&n, sizeof(int), 1, f);
	fread(&w, sizeof(int), 1, f);

	test->n = n;
	test->w = w;

	if( (test->ex = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (test->Y = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	if( (test->results = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	fread(test->ex, sizeof(real), n*w, f);
	fread(test->Y, sizeof(real), n, f);

	fclose(f);
}

void loadBinWTEST(TEST *test, char *fc)
{
	FILE *f;
	int n, w;

	if( (f = fopen(fc, "rb")) == NULL)
		error();

	fread(&n, sizeof(int), 1, f);
	fread(&w, sizeof(int), 1, f);

	test->n = n;
	test->w = w;

	if( (test->ex = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (test->results = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	fread(test->ex, sizeof(real), n*w, f);

	fclose(f);
}

void loadAsciiTEST(TEST *test, char *fc)
{
	FILE *f;
	int n, w;
	int i, j;

	if( (f = fopen(fc, "rt")) == NULL)
		error();

	fscanf(f, "%d", &n);
	fscanf(f, "%d", &w);

	w -= 1;
	test->n = n;
	test->w = w;

	if( (test->ex = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (test->Y = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	if( (test->results = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	for(i = 0; i < n; i++)
	{
		for(j = 0; j < w; j++)
#ifdef USEDOUBLE
			fscanf(f, "%lf", &test->ex[i*w+j]);
#else
			fscanf(f, "%f", &test->ex[i*w+j]);
#endif

#ifdef USEDOUBLE
		fscanf(f, "%lf", &test->Y[i]);
#else
		fscanf(f, "%f", &test->Y[i]);
#endif

	}

	fclose(f);
}

void loadAsciiWTEST(TEST *test, char *fc)
{
	FILE *f;
	int n, w;
	int i, j;

	if( (f = fopen(fc, "rt")) == NULL)
		error();

	fscanf(f, "%d", &n);
	fscanf(f, "%d", &w);

	test->n = n;
	test->w = w;

	if( (test->ex = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (test->results = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	for(i = 0; i < n; i++)
	{
		for(j = 0; j < w; j++)
#ifdef USEDOUBLE
			fscanf(f, "%lf", &test->ex[i*w+j]);
#else
			fscanf(f, "%f", &test->ex[i*w+j]);
#endif
	}

	fclose(f);
}

void setKernel(void)
{
	switch(t_kernel)
	{
	 case 0 :
		 kernel = kernel_linear;
		 break;
	 case 1 :
		 kernel = kernel_poly;
		 break;
	 case 2 :
		 kernel = kernel_exp;
		 break;
	 case 3 :
		 kernel = kernel_sigm;
		 break;
	 case 4 :
		 kernel = kernel_user;
		 break;
	 default :
		 printf("\n\nWhat's this kernel ????\n\n");
		 exit(0);		 
	}
}

void createModelFileAsciiSVM(SVM *svm, char *fc)
{
	FILE *f;
	int i, j, n, w;
	char truc[80];

	if( (f = fopen(fc, "rt")) == NULL)
		error();

	fgets(truc, 80, f);
	fscanf(f, "%d", &regression);
	fscanf(f, "%d", &t_kernel);
	fscanf(f, "%d", &k_poly);
#ifdef USEDOUBLE
	fscanf(f, "%lf", &k_exp);
	fscanf(f, "%lf", &k_sigm_s);
	fscanf(f, "%lf", &k_sigm_c);
	/* The first one is to get the end line */
	fgets(truc, 80, f);
	fgets(truc, 80, f);
	if( (k_user = malloc(sizeof(char)*strlen(truc))) == NULL )
		m_error();
	strcpy(k_user, truc);
	fscanf(f, "%d", &svm->n);
	fscanf(f, "%d", &svm->w);
	fscanf(f, "%lf", &svm->b);
#else
	fscanf(f, "%lf", &k_exp);
	fscanf(f, "%lf", &k_sigm_s);
	fscanf(f, "%lf", &k_sigm_c);
	/* The first one is to get the end line */
	fgets(truc, 80, f);
	fgets(truc, 80, f);
	if( (k_user = malloc(sizeof(char)*strlen(truc))) == NULL )
		m_error();
	strcpy(k_user, truc);
	fscanf(f, "%d", &svm->n);
	fscanf(f, "%d", &svm->w);
	fscanf(f, "%lf", &svm->b);
#endif

	n = svm->n;
	w = svm->w;

	if( (svm->points = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (svm->alpha = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();
	
	for(i = 0; i < n; i++)
	{
#ifdef USEDOUBLE
		fscanf(f, "%lf", &svm->alpha[i]);
		for(j = 0; j < w; j++)
			fscanf(f, "%lf", &svm->points[i*w+j]);
#else
		fscanf(f, "%f", &svm->alpha[i]);
		for(j = 0; j < w; j++)
			fscanf(f, "%f", &svm->points[i*w+j]);
#endif

	}

	fclose(f);
}

void createModelFileBinSVM(SVM *svm, char *fc)
{
	FILE *f;
	int i, n, w;
	char truc[80];

	if( (f = fopen(fc, "rb")) == NULL)
		error();


 	fread(&i, sizeof(int), 1, f);
	fread(truc, sizeof(char), i, f);
	fread(&regression, sizeof(int), 1, f);
	fread(&t_kernel, sizeof(int), 1, f);
	fread(&k_poly, sizeof(int), 1, f);
	fread(&k_exp, sizeof(real), 1, f);
	fread(&k_sigm_s, sizeof(real), 1, f);
	fread(&k_sigm_c, sizeof(real), 1, f);
 	fread(&i, sizeof(int), 1, f);
	if( (k_user = (char *)malloc(sizeof(char)*i)) == NULL)
		m_error();
 	fread(k_user, sizeof(char), i, f);
 	fread(&svm->n, sizeof(int), 1, f);
 	fread(&svm->w, sizeof(int), 1, f);
 	fread(&svm->b, sizeof(real), 1, f);

	n = svm->n;
	w = svm->w;

	if( (svm->points = (real *)malloc(n*w*sizeof(real))) == NULL)
		m_error();

	if( (svm->alpha = (real *)malloc(n*sizeof(real))) == NULL)
		m_error();

	fread(svm->points, sizeof(real), n*w, f);
	fread(svm->alpha, sizeof(real), n, f);

	fclose(f);
}

void checkModelSVM(SVM *svm)
{
	printf("\n# Checking SVM...\n\n");
	printf("  # NB of SV  : %d\n", svm->n);
	printf("  # dimension : %d\n", svm->w);
	printf("  # b         : %f\n", svm->b);
	printf("  # OK\n\n");	
}

real evaluateSVM(SVM *svm, real *x)
{
	real sum;
	real *ptr1, *ptr2;
	int i;

	width = svm->w;

	ptr1 = svm->points;
	ptr2 = svm->alpha;
	sum = 0;
	for(i = 0; i < svm->n; i++)
	{
		sum += *ptr2++ * (*kernel)(ptr1, x);
		ptr1 += width;
	}
	
	sum += svm->b;

	return(sum);
}
